import java.util.*;

class Clause {
    List<Predicate> predicates;

    Clause(List<Predicate> predicates) {
        this.predicates = predicates;  // connected with OR
    }
	
	void setPredicates(List<Predicate> newPredicates){ // deep set
		this.predicates = new ArrayList<>();
		for(Predicate p: newPredicates){
			List<Term> terms = new ArrayList<>();  // deep copy of terms list
			for (Term t: p.terms){
				terms.add(new Term(t.name, t.isVariable));
			}
			this.predicates.add(new Predicate(new String(p.name), terms, p.isNegated));
		}
	}
	
	private List<Predicate> copyPredicates(List<Predicate> predicates){
		List<Predicate> newPredicates = new ArrayList<>();
		for(Predicate p: predicates){
			List<Term> terms = new ArrayList<>();  // deep copy of terms list
			for (Term t: p.terms){
				terms.add(new Term(t.name, t.isVariable));
			}
			newPredicates.add(new Predicate(new String(p.name), terms, p.isNegated));
		}
		return newPredicates;
	}

    @Override
    public boolean equals(Object obj) {
        if (this == obj) return true;
        if (obj == null || getClass() != obj.getClass()) return false;
        Clause other = (Clause) obj;
		List<Predicate> p1 = this.copyPredicates(this.predicates);
		List<Predicate> p2 = this.copyPredicates(other.predicates);
		Collections.sort(p1);
		Collections.sort(p2);
        return Objects.equals(p1, p2); // Compare "sorted" lists of predicates
    }

    @Override
    public int hashCode() {
        return Objects.hash(new HashSet<>(predicates));
    }

    @Override
    public String toString() {
        return predicates.toString();
    }
}
