import java.util.*;

class Resolver {
    List<Clause> knowledgeBase;
	Unifier unifier;

	Resolver(){
		this.knowledgeBase = new ArrayList<>();
		this.unifier = new Unifier();
	}

    void addClause(Clause clause) {
        this.knowledgeBase.add(clause);
    }
	
	boolean resolve(Clause goal){
		List<Clause> toResolve = new ArrayList<>();
		toResolve.addAll(this.knowledgeBase);
		toResolve.add(goal);  // add negated goal
		System.out.println("Initial KB: " + toResolve);
		
		while(true){
			
			ArrayList<Clause> newClauses = new ArrayList<>();
			
			
			for(int i = 0; i < toResolve.size(); i++)  // updated at the end with new clauses (if any)
            {
                Clause Ci = toResolve.get(i);

                for(int j = i+1; j < toResolve.size(); j++)
                {
                    Clause Cj = toResolve.get(j);
					
					System.out.println("\nResolving " + Ci + " with " + Cj);
					List<Clause> resolvedClauses = this.resolveClauses(Ci, Cj);
					
					for (Clause resolvedClause: this.emptyIfNull(resolvedClauses)){
						if (resolvedClause != null && resolvedClause.predicates.isEmpty()) {
							// Contradiction found: Empty clause
							System.out.println("\nResolving " + Ci + " with " + Cj + " ended up on empty clause!\n");
							return true;
						}
	
						if (resolvedClause != null && !toResolve.contains(resolvedClause) && !newClauses.contains(resolvedClause)){
							// Add only new clauses
							
							newClauses.add(resolvedClause);  // with changed variables!!
						}
					}
					
				}
			}
			
			boolean newClausesFound = false;
            //Check if any new clauses were produced in this loop
            if(newClauses.size() > 0)
            {
            	toResolve.addAll(newClauses);
                newClausesFound = true;
            }
			
			//change variables in updated KB
			for(Clause c: toResolve){
				Set<Predicate> changedPredicates = this.changeVariables(new HashSet<>(c.predicates));
				c.setPredicates(new ArrayList<>(changedPredicates));
			}
			System.out.println("Updated KB: " + toResolve);
            

            //If not, then Knowledge Base does not logically infer the literal we wanted to prove
            if(!newClausesFound)
            {
                System.out.println("New clauses were not found.");
                return false;
            }
			
		}
	}


    private List<Clause> resolveClauses(Clause c1, Clause c2) {
		List<Clause> newClauses = new ArrayList<>();
        for (Predicate p1 : c1.predicates) {
            for (Predicate p2 : c2.predicates) {
                if (p1.isNegation(p2)) {
                    boolean canUnify = false;
                    for (int i = 0; i < p1.terms.size(); i++) {
						System.out.println("Trying to unify predicates " + p1 + " and " + p2 + "\n");
						
                        if (this.unifier.unify(p1.terms.get(i), p2.terms.get(i))) {
                            canUnify = true;
                            break;
                        }
                    }
                    if (canUnify) {
                        Set<Predicate> newPredicates = new HashSet<>();
                        newPredicates.addAll(c1.predicates);
                        newPredicates.addAll(c2.predicates);
                        newPredicates.remove(p1);
                        newPredicates.remove(p2);
						System.out.println(c1 + " and " + c2 + " resolved with unifier: " + this.unifier.substitutions);
						Set<Predicate> changedPredicates = new HashSet<>();
						for (Predicate p: newPredicates){
							changedPredicates.add(this.changeVariables(p)); // method for single Predicate
						}
						// Set<Predicate> changedPredicates = this.changeVariables(newPredicates);
						// System.out.println("Adding predicates " + changedPredicates + " to KB");
                        // return new Clause(new ArrayList<>(changedPredicates));
						newClauses.add(new Clause(new ArrayList<>(changedPredicates)));  // empty clause can be added here...
                    }
                }
            }
        }
		if (newClauses.size() > 0) {  // if resolution applied
			System.out.println("Adding predicates " + newClauses + " to KB");
			return newClauses;
		}
		System.out.println("Not resolved! No opposite terms for resolution rule...\n");
        return null;
    }
	
	private Set<Predicate> changeVariables(Set<Predicate> predicates){
		Set<Predicate> changedPredicates = new HashSet<>();  // deep copy of input predicates
		// deep copy procedure 
		for (Predicate p: predicates){
			List<Term> terms = new ArrayList<>();  // deep copy of terms list
			for (Term t: p.terms){
				terms.add(new Term(t.name, t.isVariable));
			}
			changedPredicates.add(new Predicate(new String(p.name), terms, p.isNegated));
		}
		for (Predicate p: changedPredicates){
			// System.out.println(p);
			for (int i = 0; i < p.terms.size(); i++){
				// System.out.println(p.terms.get(i).name);
				if (this.unifier.substitutions.containsKey(p.terms.get(i).name)){  // String comparison
					System.out.println("Changing " + p.terms.get(i).name + " with " + this.unifier.substitutions.get(p.terms.get(i).name) + " in predicate " + p);
					boolean var = true;
					if (!this.unifier.substitutions.get(p.terms.get(i).name).contains(Character.toString('x'))) var = false;
					// p.terms.set(i, new Term(this.unifier.substitutions.get(p.terms.get(i).name), p.terms.get(i).isVariable));
					p.terms.set(i, new Term(this.unifier.substitutions.get(p.terms.get(i).name), var));
				}
			}
		}
		return changedPredicates;
	}
	
	private Predicate changeVariables(Predicate p){
		// deep copy of Predicate
		List<Term> terms = new ArrayList<>();  // deep copy of terms list
		for (Term t: p.terms){
			terms.add(new Term(t.name, t.isVariable));
		}
		Predicate newPredicate = new Predicate(new String(p.name), terms, p.isNegated);
		for (int i = 0; i < newPredicate.terms.size(); i++){
			if (this.unifier.substitutions.containsKey(newPredicate.terms.get(i).name)){
				System.out.println(
					"Changing " + newPredicate.terms.get(i).name + " with " + this.unifier.substitutions.get(newPredicate.terms.get(i).name) + " in predicate " + newPredicate
				);
				boolean var = true;
				if (!this.unifier.substitutions.get(newPredicate.terms.get(i).name).contains(Character.toString('x'))) var = false;
				// newPredicate.terms.set(i, new Term(this.unifier.substitutions.get(newPredicate.terms.get(i).name), newPredicate.terms.get(i).isVariable));
				newPredicate.terms.set(i, new Term(this.unifier.substitutions.get(newPredicate.terms.get(i).name), var));
			}
		}
		return newPredicate;
	}
	
	private <T> Iterable<T> emptyIfNull(Iterable<T> iterable) {
		return iterable == null ? Collections.<T>emptyList() : iterable;
	}
}